/*
 *	Serial port driver for 16Cxx chips
 *	using software delays.
 *
 *	Copyright (C)1996 HI-TECH Software.
 *	Freely distributable.
 */

/*
 *	Tunable parameters
 */

/*	Transmit and Receive port bits */
#include "common.h"

#define TxData RA3
#define RxData RA4

/*	Xtal frequency */
// Note with V7.86pl4 of the compiler xx.xx will NOT WORK!
#ifndef XTAL
#define	XTAL	10000000
#endif

/*	Baud rate	*/
#define	BRATE	4800

/*	Don't change anything else */
#define	DLY		3		/* cycles per null loop */
#define	TX_OHEAD	13		/* overhead cycles per loop */
#define	RX_OHEAD	12		/* receiver overhead per loop */

#define	DELAY(ohead)	(((XTAL/4/BRATE)-(ohead))/DLY)

//Send 1start +8bits + 1stop
//Operation is INVERTED!
void putch(unsigned char c)
{
 unsigned char dly, bitno;

 bitno = 11;

 TxData = 1; /* start bit */
 do
  {
   dly = (unsigned char)DELAY(TX_OHEAD); /* wait one bit time */
   while(--dly);
   if(c & 1)
    TxData = 0; //Inverted
   if(!(c & 1))
    TxData = 1; //Inverted
   c = (c >> 1) | 0x80;
  } while(--bitno);
 TxData = 0; //Force idle, inverted
}

//Receive 1start +8bits +1stop
//Operation is INVERTED!
unsigned char getch(void)
{

 unsigned char c, bitno, dly;

 for(;;)
  {

/****************************************************************************
  If MIC-E is active the PTT is polled here.

  When the PTT is deactivated the send_mice flag will be set and the
  function will return.

  The ptt will be activated here.
  De-keying will be done in the main loop.
*/
   while(!RxData)
    {

     if ( !sw2 )
      {

       //Is PTT active?
       if ( !ptt_in )
	{

	 //Not allready keyed up?
	 if ( !ptt_active )
	  {

	   //Need to Key up
	   ptt_out = 1;
	   ptt_active = 1; //Signal PTT active

	  }
	}
       else
	{

	 //PTT not active

	 //Was it active?
	 if ( ptt_active )
	  {

	   send_mice = 1; //Signal main loop
	   return 0; //Return to send the beacon

	  }
	} //End ptt_in
      }
     else if ( ptt_active ) //sw2
      {

       //Fix lockup when changing from MIC-E to beacon mode while the
       //PTT is still de-pressed
       ptt_out = 0;
       ptt_active = 0;

      } //End sw2
    } //Wait for start bit
   dly = (unsigned char)DELAY(3)/2; //Do oversample
   while(--dly);
   if(!RxData) continue; /* twas just noise */
   bitno = 8; //Get 8 bits
   c = 0;
   do
    {
     dly = (unsigned char)DELAY(RX_OHEAD);
     while(--dly); //Wait for bit
     c = (c >> 1) | (RxData << 7); //Sample and add to byte
    } while(--bitno);

   //Wait for stopbit!!!!!
   dly = (unsigned char)DELAY(RX_OHEAD);
   while(--dly); //Wait for bit

   return (unsigned char)~c; //Return Data INVERTED!
  }
}
