//
//    ***********************************************************************
//    *                        COPYRIGHT NOTICE                             *
//    *      (c) Copyright October 2000, Michael Pendley, K5ATM             *
//    *                                                                     *
//    * This software is available for non-commercial use only. Use by      *
//    * amateur radio operators and applications associated with education  *
//    * permitted without fees.  The author retains all rights to all       *
//    * other applications.  Contact the author for commercial applications *
//    ***********************************************************************
//
//
// SOTT --- Son of TinyTrak
//
// Revision History
//
//    10/1/2000
//
//       Version 1.0 started
//
//
//
//  How it came to be
//       Byon Garrabrant's TinyTrak is one of the coolest kits I have
//       built in a long time -- simply because of its minimilistic
//       approach.
//
//       While the original TinyTrak worked just fine and all, I felt
//       compelled to try and make it better.  For example, the original
//       software did not support the notion of transmit hold off when
//       GPS data was invalid. Interestingly enough -- neither did the
//       code GPS-E-2 by John Hansen, W2FS -- the standard code used to
//       drive the TAPR PIC-E).  GPS-E-2 is available from www.tapr.org
//
//       SOTT is usefull to non APRS types as well.  Users wishing
//       to broadcast general packet data can delete the routines
//       related to the GPS and keep the the general routines such
//       as the tone generating routine, the sendbyte routine, the CRC
//       generating routine, etc.  All one would need to add is a routine
//       to acquire the data, format it, then send out each byte with
//       sendbyte.
//
//
//  How is SOTT different (as an APRS tracker)?
//
//       I decided the eliminate the MIC-e format simply because it is
//       complicated -- and complicated formats require a lot of program
//       space.  If you want Mic-e, I suggest you look at W2FS's code.  He
//       has a nice implementation.  Packets are not a lot bigger than
//       Mic-e (especially after you consider the total packet size with
//       flags) and its simpler structure used less program space.
//
//       SOTT is different in that
//          1) GPS time is used to time transmissions.  This gives the
//             user very fine control of transmissions but has the side
//             effect of not allowing beacons to be transmitted if GPS
//             data is unavailable.
//          2) Transmission is held off until GPS data is valid.  This
//             includes beacons.
//          3) A heart beat was added to let the user know the GPS is
//             is alive.  A blinking valid LED says data from the GPS
//             is being received.  A blink that is mostly on says the data
//             is valid.  A blink that is mostly off says the data is
//             not valid.
//          4) PTT is not hit at start up time
//
//
//  General comments
//
//       The most difficult code to develop was the interrupt routine
//       that generates the 1200Hz and 2200Hz tones.  See that section
//       of the code for details.
//
//       Source code is written in PCW C compiler.  A listing file is
//       provided should a dedicated hacker without a C compiler wish
//       to cut and paste the generated code into the MPLAB assembler.
//
//       Even though it will be hard to find much of the original
//       code I feel obligated to admit that I used GPS-2-E as a guide
//       to creating this program (and even lifted a couple of utility
//       functions).

/****************************************************************************
 Modifications done by Brett ZL1TFA.

 Ported the code to the Hi-Tech PIC-L compiler, this means it can be
 modified by the free version avaliable for download from thier web site
 at http:\\www.htsoft.com

 I have include 2 project files. One for the compiler and the other for
 mp-lab.
 You will need to edit the location of the source files etc..

 Added code to support a sort of MIC-E mode.
 This is controlled by SW2. When SW2 is off the tracker works as normal.
 When SW2 is on the beacon stops and data is sent when the PTT switch is
 released.
 The ptt_out line is controlled by the PIC and the microphone PTT MUST be
 connected to the ptt_in line on the pic.

 If the GPS data is invalid while in MIC-E mode then data transmited will
 be the last valid received GPS data.

 While MIC-E mode is active no beacon message will be sent.

 If MIC_E is defined in common.h the packet data will be sent in mic-e format
 just like the original TinyTrak.

 I have found a problem with the ptt_in line on the TinyTrak design where
 there is not enough current to insure forward conduction of the diode in
 all cases. This can be fixed by adding a 470ohm resistor between pin 11 and
 the 5V line.

 I also STRONGLY recomend switching the PTT line of the radio between the
 pic and the microphone in case of any problems or you wish to disable the
 packet for important overs.
 Also it may be necessary to isolate the microphone audio line while the
 packet is being sent as it will still be live and any background noise
 may cause disruption of the packet data being sent.

 I have been using crystals, not the resonator, as in the original design,
 as they have far too much drift in changing temperatures!

 **WARNGING** IF THE GPS DATA IS DISRUPTED IN ANY WAY THE PTT WILL REMAIN
	      LOCKED ON!!
	      ALSO IF YOUR GPS DOES NOT SUPPLY A STEADY STREAM OF DATA THE
	      PTT WILL BE SLOW RESPONDING!
	      This problem will only occur in MIC-E mode!
	      This is only beacuse I am polling the ptt_in line at the START
	      of a message from the GPS.

 On the matter of the timing for the tone generating I have changed the loop
 count values from the original code as the packet data WAS not being decoded
 by all TNCs!

 IT WILL BE NECESSARY TO CHECK THE TIMING IF YOU MAKE ANY CHANGES TO THE
 PACKET ROUTINES OR IF YOUR COMPILER HAS DIFFERENT OPTIMIZATIONS.
 IN OTHER WORDS CHECK IT BEFORE YOU USE IT!

 I am of the opinion that the loop count values MUST be a user settable
 paramater and hence I will look at incorperating it.
 Also I will be changing the communication for config to use echo for
 handshaking purposes.

 Version 2.0
 -----------
 No MIC-E data format. Also had bug with beacon rate.

 Version 2.1
 -----------
 Bug fixes.
 Added MIC-E data format.

 Version 2.2
 -----------
 Added the option for the units callsign to be sent as a morse code ID at the
 end of each packet transmission.

 This will only occur in BEACON mode, if the unit is placed into MIC-E mode
 the morse ID WILL NOT BE SENT!

 **HAS NOT BEEN TESTED**

 Version 2.3
 -----------
 Morse ID is fixed at 15mins.
 A morse ID will always be sent at power up, if used.

 Version 2.4
 -----------
 Fixed the morse routine not sending the number 9.
 The morse routine was doing the element pause after the last element was
 sent.
 The mic-e / beacon switch would lock the PTT if changed from mic-e to
 beacon while keyed.
 Added the TIMED_MIC_E option.
 Added the No GPS message for the MIC-E mode.

 Version 2.5
 -----------
 Added altitude for the non-compressed mode. Just takes the RAW data from the
 GGA sentance and formats it for sending!

 Version 2.6
 -----------
 Some bug fixes.
 Changed the eeprom.as file & the CONFIG macro for the new PICC8.xx format.

 Version 2.7
 -----------
 Removed the altitude as it was in meters and not feet!
*/

#define MAIN
#include "common.h"
#include <ctype.h>

#ifdef MORSE
/****************************************************************************
 Send tabel for A - Z

 Format of the table entries:

 Bits 7 - 3 are the characters element where a 1 is a dash and a 0 is a dot.

 Bits 2 - 0 count of the elements.
*/
const unsigned char a_tabel[] =
 {

  66,  /* A */
  132, /* B */
  164, /* C */
  131, /* D */
  1,   /* E */
  36,  /* F */
  195, /* G */
  4,   /* H */
  2,   /* I */
  116, /* J */
  163, /* K */
  68,  /* L */
  194, /* M */
  130, /* N */
  227, /* O */
  100, /* P */
  212, /* Q */
  67,  /* R */
  3,   /* S */
  129, /* T */
  35,  /* U */
  20,  /* V */
  99,  /* W */
  148, /* X */
  180, /* Y */
  196  /* Z */

 };

/****************************************************************************

 Send tabel for 0 - 9.

 Foramt same as above.
*/
const unsigned char n_tabel[] =
 {

  253, /* 0 */
  125, /* 1 */
  61,  /* 2 */
  29,  /* 3 */
  13,  /* 4 */
  5,   /* 5 */
  133, /* 6 */
  197, /* 7 */
  229, /* 8 */
  245  /* 9 */

 };
#endif //MORSE

//Setup CONFIG
//HS,no code protection
__CONFIG(XT&UNPROTECT&PWRTEN&WDTDIS);

//Setup IDLOC
__IDLOC(0000);

///////////////////////////////////////////////
//                                           //
//             Interrupt Routine             //
//                                           //
///////////////////////////////////////////////
void interrupt tone(void)
{

/*
    Tone generation is accomplished with this interrupt routine.
    Interrupts are turned off when RF is off.

    Once an APRS message is ready to transmit, interrupts are turned on
    tone generation starts.  Up to 87% of the CPU is consumed
    during tone generation.  The remaining 13% is used to process
    process bits (e.g. get the next byte to send, strip out bits,
    change tone when needed, bit stuff, etc.  All times assume a
    10MHz clock.

    OK lets get started.  A timr0 interrupt just happened and control
    has been transferred here.  It took 2 instruction cycles (.8us) to
    process the request.  These two instruction cycles are accounted for
    with the "dummy" cycle NOPs below.  The times to the left of the
    instructions indicate how much time has passed until this instruction
    STARTS.

    This routine assumes that the following steps were taken BEFORE
    interrupts were enabled:

     1) set "base" to the nominal timer0 value (tone1200 or tone2200)
     2) initialize dink to the 1200 or 2200 Hz value
     3) initialize r-val to 0 (first time through)
     4) loop_ctr initialized as needed

*/
#asm

//      nop                   // 0.0    dummy cycle, used as place holder
//      nop                   // 0.4    dummy cycle, used as place holder
	movwf   _gpsdata      //0.8us   Save state
	swapf   _STATUS,w     //1.2us
	movwf   _gpsdata+1    //1.6us

	movf    _base,w       //2.0us   Lets reset the timer first

	decfsz  _dink,f       //2.4us   Do we need to extend the loop time?
	goto    nodink        //2.8us

	addlw   0xff          //3.2us   Yes ... W = W + (-1)
	movwf   _TMR0         //3.6us   At the end of this instruction
			      //        timer0 is one less than the value
			      //        of timr0 at the end of the 3.6us
			      //        instruction below.
	movf    _gpsdata+2,w  //4.0us   Now lets update the sine wave
	call    getsin        //4.4us   Get wave profile from table
	movwf   portb         //6.4us   portb is updated at 6.4us in both places

	movf    _gpsdata+5,w  //6.8us   we need to restore (reinit) dink
	movwf   _dink         //7.2us

	goto    incloop       //7.6us

nodink:
	movwf   _TMR0         //3.6us   No loop time dink needed
	movf    _gpsdata+2,w  //4.0us   Now lets update the sine wave
	call    getsin        //4.4us
	movwf   portb         //6.4us


// We now have two running times.

// The first value to the right of the instruction indicates when that
// instruction starts if no loop dink was needed.
//
// The second value indicates when that instruction starts if a loop dink
// was required.
//
// However, since the timer and sine wave were updated at the same time
// regardless of which path we take (3.6us and 6.4us), the diverging times
// here will not effect the sine wave frequency.  We only need to keep
// track of the time to be sure we get out of the interrupt before the
// next loop is due to begin (i.e. timr0 rolls to 0xff), and so we can
// calculate processor load (see below).
//

incloop:

	incf    _gpsdata+2,w  //8.4-6.8  set pointer to next sine wave value
	andlw   0x1f          //8.8-7.2  0,1,2,...,31,0,1,...
	movwf   _gpsdata+2    //9.2-7.6

	incf    _gpsdata+3,f  //9.6-8.0   Inc loop_ctr -- used to time bit length

	swapf   _gpsdata+1,w  //10.0-8.4  Restore state and return
	movwf   _STATUS       //10.4-8.8
	swapf   _gpsdata,f    //10.8-9.2  restore W (without changing status)
	swapf   _gpsdata,w    //11.2-9.6

	bcf     0x0b,2        //11.6-10

	retfie                //12-10.4
//      nop                   //12.4-10.8  dead cycle during return from interrupt
//      nop                   //12.8-11.2  dead cycle during return from interrupt
//
//      <next instruction>    //13.2-11.6

//  Worse case is 2200Hz. If we dinked, timr0 should expire at 14.4us
//  so we have 1.2us of free time (3 instructions)
//  If we did not dink, timr0 should expire at 14us so we have 2.4us of
//  free time (6 instructions)
//
//  Each bit is about 32 loops so we have between 96 and 192 instructions
//  to process the next bit.  If we set dink22 to 2, we dink every other
//  loop so we have 16*3 + 16*6 = 144 instructions to process the next
//  bit.  This represents a processor load of about 12.4us/14.2us = 87.3%
//  giving us a processor with an effective clock of 1.27MHz.
//
//  An analysis of the 1200Hz case yields a load of about 48%
//
//
//  I have not performed a detailed analysis of bit and byte processing
//  times while transmitting but analysis with an o-scope indicates that
//  control always reaches the wait loops in sendbyte well before the bit
//  timer (loop_ctr) expires.
//
//
// Here is the sine wave profile.  The indicated voltages are measured
// from the top of R5 and assume a nominal 5 volts out of the pic chip
// I/O lines
getsin:
	addwf   02,f
//                       r-val   volts
	retlw   0x7    //  0      2.33
	retlw   0x6    //  1      2
	retlw   0x5    //  2      1.67
	retlw   0x4    //  3      1.33
	retlw   0x3    //  4      1
	retlw   0x2    //  5      .67
	retlw   0x1    //  6      .33
	retlw   0x1    //  7      .33
	retlw   0x1    //  8      .33
	retlw   0x1    //  9      .33
	retlw   0x2    // 10      .67
	retlw   0x3    // 11      1
	retlw   0x4    // 12      1.33
	retlw   0x5    // 13      1.67
	retlw   0x6    // 14      2
	retlw   0x7    // 15      2.33
	retlw   0x9    // 16      3
	retlw   0xa    // 17      3.33
	retlw   0xb    // 18      3.67
	retlw   0xc    // 19      4
	retlw   0xd    // 20      4.33
	retlw   0xe    // 21      4.67
	retlw   0xf    // 22      5
	retlw   0xf    // 23      5
	retlw   0xf    // 24      5
	retlw   0xf    // 25      5
	retlw   0xe    // 26      4.67
	retlw   0xd    // 27      4.33
	retlw   0xd    // 28      4.33
	retlw   0xb    // 29      3.67
	retlw   0xa    // 30      3.33
	retlw   0x9    // 31      3

#endasm
} // tone

// So how do we decide what values to set tone1200 and tone2200 to?
//
// (1) Timr0 is incremented near the end of an instruction cycle.
// (2) The interrupt is tested near the beginning of an instruction cycle
// (3) The timr0 increment is inhibited for two inst cycles after a write
//
// So .. assume we write 0xff to timr0 and the timr0 interrupt is enabled.
//
//          action              start-end        comment
//
//    Write 0xff to timr0       0.0-0.4us  timr0 is ff at instruction end
//    next inst is executed     0.4-0.8us  timr0 is inhibited (still ff)
//    next inst is executed     0.8-1.2us  timr0 is inhibited (still ff)
//    next inst is executed     1.2-1.6us  timr0 is 00 at end of instruction but ff when tested
//                                         (tested at beginning of instruction, incremented @ end)
//    next inst is executed     1.6-2.0us  timr0 test yields 0 and interrupt will kick in at end of instruction
//    dummy cycle               2.0-2.4us
//    dummy cycle               2.4-2.8us
//    1st int. inst. executed   2.8-3.2us
//
// So if, at the first instruction above, we say -- I want to start my first
// dummy interrupt cycle in 2.0 us -- which is 5 instructions (5*.4 = 2.0) -- I am
// tempted to put 0xff-5 into timr0.  However, we see from above we really need
// to put in 0xff-0.
//
// For this application 00000000
//   We want interrupts to be 14.2045us apart for 2200 Hz
//      (1/2200)/32 = 14.2045us
//   We want interrupts to be 26.04166us apart for 1200 Hz
//      (1/1200)/32 = 26.04166
//
// Because we use integer arithmetic, we can set the 2200Hz interrupt
// rate to 14us (35*.4) and the 1200Hz interrupt rate at 26us (65*.4)
// these rates will make the tones run slightly fast.  We will slow down
// the times by adding one clock cycle (dinking) individual 1/32 of a
// cycle subloops.
//
// So what values do we use for the counters?  For 2200Hz we need 35
// clock cycles.  We loose 5 do to the way the F84 works. Looking at the
// interrupt routine, we consume 9 in processing the first part of the
// interrupt.  This means we need to set timr0 to 0xff-21 (35-14=21).
// For the 1200Hz tone we use 0xff-51 (65-14 = 51).  These are of course
// nominal values may need to be adjusted depending on the actual clock frequency
//
// SendByte has a requirement to wait 833-860us (1 bit time) before processing
// the next bit.  We will use the interrupt loop to keep this time.  If
// the tone is set to 2200Hz (14.2us / average loop) we will preset
// loop_ctr to 833/14.2=58.66  or 860/14.2=60.5.  59 is in the middle.
//
// For 1200Hz we have 833/26.04 = 31.9 or 860/26.04 = 33.  My experiments
// show 33 worked best for me.
//
// To get the frequencies exact, every once and a while we add 1 clock
// cycle to one of the 32 sub-loops required to make one tone cycle.  For
// example, assuming a clock rate of exactly 10MHz, a normal 2200Hz
// loop is 1/(35*32*.4) = 2232.14Hz.  If we add .4us to one of the
// sub-loops we have 1/(.4+(35*32*.4)) = 2230.15Hz.
//
// To implement this, the register "dink" is decremented every time
// the interrupt is entered. If dink=0 it is reset back to the base dink
// value and the time to the next interrupt is extended by one clock cycle.
// Example -
//   If the base dink value is 1 then every cycle is extended
//     1->0 (extended, reset dink to 1)
//   If the base dink value is 2 then every other cycle is extended
//     2->1 (normal)   1->0 (extended, reset dink to 2)
//   If the base dink value is 3 then every third cycle is extended
//     3->2(normal) 2->1(normal) 1->0 (extended, reset dink to 3)
//
// We end up with a non-linear relationship between the final
// frequency and the value of dink.  That relationship is approximately:
//     For 1200Hz, final frequency = 1/((32/D)*.4+832)
//     For 2200Hz, final frequency = 1/((32/D)*.4+448)
//
// For the 1200 Hz region, changing dink from 1 to 2 will change the
// the final frequency by about 9Hz.  Changing dink from 13 to 14
// however, will only change the final frequency by about 0.1Hz
//
//
// Bottom line: You ought to be able to find values for the course and
// fine frequency controls to get within 0.5 percent of a
// desired frequency.
//
//
///////////////////////////////////////////////
//                                           //
//                   Init                    //
//                                           //
///////////////////////////////////////////////
//
// A simple init routine but since it is called twice we
// saved a little program space by making it a subroutine
// and speed is not an issue in this case.
//
//  Called by
//     config
//     main
void init (void)
{

 validpos  = 0;
 firstTime = 1;
 beacon = 1;        // first xmit will include a beacon

} // init

///////////////////////////////////////////////
//                                           //
//                 flashcd()                 //
//                                           //
///////////////////////////////////////////////
//
//  Update the state of the carrier detect lamp
//
//  Called by
//    readgps
//    getposition
//    main
void flashcd(void)
{

 // Set inverted state of the pin
 cdflag = (audio ? 0 : 1);
 cd = cdflag; //Update LED

}
//flashcd

///////////////////////////////////////////////
//                                           //
//                flipout()                  //
//                                           //
///////////////////////////////////////////////
//
// This routine will cause the "other" tone to be
// generated.  Example, if 1200Hz tone is in progress
// then the next tone will be 2200Hz. We do this
// by changing the base interrupt delay value.
//
//    Called by
//       config
//       sendbyte
void flipout(void)
{

 stuff = 0; //flipping means we are sending a zero so reset the 1s counter
 if(doing1200)
  {
   doing1200 = 0;
   base = cbase2200;
   dink = base_dink = 2; //Hard coded dink for 2200Hz
  }//if
 else
  {
   doing1200 = 1;
   base = cbase1200;
   dink = base_dink = 1; //Hard coded dink for 1200Hz
  } // else
}// flipout

///////////////////////////////////////////////
//                                           //
//                initTone()                 //
//                                           //
///////////////////////////////////////////////
//
// This subroutine will init all state needed to
// start generating a tone.  1200Hz is set as the
// sefault.  If 2200Hz is desired, a call to flipout
// should be made after initTone.
//
// Called by
//   config
//   sendpacket
//
// Modified to allow the loop count values for tone length to bs user set.
// In order to be compatable with the PC software the fine tune paramaters
// have been used. This means the fine tune paramters are now hard coded!
void initTone(void)
{

 dink = base_dink = 1; //Hard coded dink for 1200Hz
 base = cbase1200 = eeprom_read(EEtone1200);
 cbase2200 = eeprom_read(EEtone2200);
 loop12 = eeprom_read(EEdink12); //Tone length for 1.2KHz
 loop22 = eeprom_read(EEdink22); //Tone length for 2.2KHz
 doing1200 = 1;
 r_val   = 0;

} // initTone

///////////////////////////////////////////////
//                                           //
//                  mygetc                   //
//                                           //
///////////////////////////////////////////////
//
// A support routine for config.  This routine
// will get a character from the serial port and
// map values as needed.
//
//   254  => 0
//   253  => 253
// ...
//   1    => 1
//
// This ugly hack is used because I cannot get Visual
// Basic to send a 0 or 0xff out the serial port. I hope
// to solve this some day.  When I do I will only need to
// cut out this routine and convert calls to mygetc back to
// getc
unsigned char mygetc(void)
{

 unsigned char character;

 character = getch();
 if (character == 254) character = 0;
 return character;

}

///////////////////////////////////////////////
//                                           //
//                 config()                  //
//                                           //
///////////////////////////////////////////////
//
// Used to upload configuration parameters into the
// TT, download current config info, and generate
// continuous 1200Hz and 2200Hz tones for calibration
//
// All communication is at the same baud rate as the GPS.
//
// Format:
//
//   Who      Character           Comment
//  -------------------------------------------------
//   user         !         Put TT in config mode
//    TT          >         Ready to accept command
//   user/tt   various      User and TT communicate
//    TT          <         TT is exiting config mode
//
//
//   The following commands are supported
//
//   D  =>  Dump current config parameters
//   U  =>  Upload config parameters
//   T  =>  1200Hz test tone
//   t  =>  2200Hz test tone
//
//
//   To dump parameters:
//
//    user:  "!"   put TT in config mode
//    TT:    ">"   TT ready to accept command
//    user:  "D"   Dump command
//    TT:     xx   Prints software version number (1 byte) followed
//                 by 64 bytes of config info in the same order it is
//                 stored in the eeprom
//    tt:    "<"   TT exits config mode
//
//
//   To upload parameters:
//
//    user:  "!"   put TT in config mode
//    TT:    ">"   TT ready to accept command
//    user:  "U"   Upload command
//    TT:    "*"   TT asks for start address
//    user:   xx   user provides 1 byte of eeprom address (0-63) - see below
//    TT:    "*"   TT asks for end address
//    user:   xx   user provides 1 byte of eeprom address
//    loop TT:    "#"   TT asks for next binary data byte
//     user:   xx   user provides next binary data byte
//     --loop as needed--
//    TT:    "<"   all data received, exit config mode
//
//    I wish I could have buffered all data to be sure things are cool before updating
//    the EEPROM but there is just not enough resources to work with for that.
//
//
//   To generate test tones:
//
//    user:  "!"   put TT in config mode
//    TT:    ">"   TT ready to accept command
//    user:  "T"   user asks for 1200 Hz tone
// or
//    user:  "t"   user asks for 2200 Hz tone
//
//  TT will generate tones until some sort of activity is seen on the serial
//  line.
//  Because interrupts are on, the RS232 timing loops will be all goofed up
//  so we cannot send across an ASCII character to say stop.
//  One could get TT attention by sending across lots of characters and hope
//  one is seen.  Better is to send a "break" (a very long space).
//
//
//
//
//  Note:
//   I was not able to get Visual Basic to output 0x00 or 0xFF
//   on the serial port (can anyone help?)  So (he says with a red face)
//   I compressed the valid data range from 0-0xff to 0-0xfd and reserved
//   0xfe and 0xff. This limits user data to 0-0Xfd.  The Visual Basic
//   interface program will first map all data (that is to be sent to the TT)
//   with the values of 0xff and 0xfe to 0xfd.  The interface will
//   then map all data (to be sent to the TT) with the value 0
//   to 0xfe
//
//   The TT routine mygetc will map 0xfe to 0.
//
//   Data from the TT to Visual Basic is not mapped in any way.  Transmitting
//   0x00 and 0xff into Visual Basic seems to work just fine!
//
void config (void)
{

#ifndef DEBUG
 putch('>');   // Say I am in config mode
#endif

 //Get user input
 Ccmd = getch();

 // Dump EEPROM
 if (Ccmd == 'D')
  {
   putch(TTvers);
   for (Cloop = 0; Cloop < 64; Cloop++)
    putch(eeprom_read(Cloop));
  }
 // Update EEPROM
 else if (Ccmd == 'U')
  {
   putch('*');
   Cstart = mygetc();
   putch('*');
   Cend = mygetc(); Cend++;
   for (Cloop = Cstart; Cloop < Cend; Cloop++)
    {
     putch('#');
     Cdata = mygetc();
     EEPROM_WRITE(Cloop, Cdata);
     ei(); //Re-enable interupts
    }
  }
 //Send test tones
 else if ((Ccmd == 'T') || (Ccmd == 't'))
  {
   initTone();
   if (Ccmd == 't') flipout();
   ptt_out=1;	//Keyup
   TMR0=0xff;	//Preset timer
   T0IE = 1;	//enable timer interrupt
   flag = 1;	//Signal doing flags?
   while(!RA4);	//Wait for any char to stop
   T0IE = 0;	//disable timer interrupt
   ptt_out = 0; //Dekey
  }

#ifndef DEBUG
 putch('<');  // Say I am leaving config mode
#endif

 init();

} // config

///////////////////////////////////////////////
//                                           //
//                 readGPS()                 //
//                                           //
///////////////////////////////////////////////
//
// Read data from the serial port (up to and
// including the next comma) into gpsdata[]. If
// gpsdata[] fills before the comma then just
// throw the extra stuff away.
//
// called by getdata
//
void readGPS(void)
{

 unsigned char i;

 flashcd();	//update the carrier detect lamp
 i=0;
 gpsdata[0]=0;	// x=0; y=0; ... generates more efficient
 gpsdata[1]=0;	// code than x=y=z=0
 gpsdata[2]=0;
 gpsdata[3]=0;
 do
  {
   stuff = getch();
   if (i < GPS_buffer_size) gpsdata[i] = stuff;
   i++;
  }
 while (stuff != ',');

}// readGPS

///////////////////////////////////////////////
//                                           //
//               getGPSfloat()               //
//                                           //
///////////////////////////////////////////////
//
// Gets the whole number portion of a
// floating point variable that has
// between 1 and 3 digits.  Places the
// digits in the three bytes of cs[]
//
// This routine shamelessly lifted from
// W2FS' code and modified
//
// called by getdata
void getGPSfloat(void)
{

 readGPS();
#ifdef MIC_E
 cs[0] = cs[1] = cs[2] = 0;
#else //MIC_E
 cs[0] = cs[1] = cs[2] = '0';
#endif //MIC_E

 // if there is 1 digit before the dot
 if (gpsdata[1] == '.') cs[2] = gpsdata[0];

 // if there are 2 digits before the dot
 else if (gpsdata[2] == '.')
  {
   cs[1] = gpsdata[0];
   cs[2] = gpsdata[1];
  }

 // if there are 3 digits before the dot
 else if (gpsdata[3] == '.')
  {
   cs[0] = gpsdata[0];
   cs[1] = gpsdata[1];
   cs[2] = gpsdata[2];
  }

#ifdef MIC_E
 cs[0] &= 0x0f;
 cs[1] &= 0x0f; //strip the top nibble off all three bytes
 cs[2] &= 0x0f;
#endif //MIC_E

}// getGPSfloat


///////////////////////////////////////////////
//                                           //
//                  getdata()                //
//                                           //
///////////////////////////////////////////////
//
// This is the main data loop.  The serial input
// line is monitored.
//
// If a "!" character is seen the TT is placed in
// command mode.
//
// If a "$" is seen then a GPS string is probably
// starting and it is processed.
//
// Once valid GPS data is found, an APRS message is formatted.
//
// We do not leave this routine until we have valid GPS data
// and an APRS message is ready to transmit
//
// MIC-E encoding lifted and modified from W2FS code.
//
// Called from:
//  Main
void getdata(void)
{

 unsigned char i;

 for(;;) //Data loop
  {

   // Wait for the start of a GPS string ($) or a config request (!)
   for(;;)
    {
     flashcd(); // Keep the RF busy lamp updated

     gpsdata[0]=getch();

     //If need to send a packet do it.
     if ( send_mice ) return;

     // ! is the config command char
     if (gpsdata[0] == '!')
      {
       config();
       continue;
      }

     // Start of GPS message string
     if (gpsdata[0] != '$') continue;

     // OK, a GPS string is starting.  Look for a GPRMC string
     readGPS();

     //Is it a RMC string?
     if (gpsdata[0] == 'G')
      if (gpsdata[1] == 'P')
       if (gpsdata[2] == 'R')
	if (gpsdata[3] == 'M')
	 if (gpsdata[4] == 'C')
	 break;

    } //End wait loop

// Looking good so far.  Lets review what a $GPRMC string looks like
// $GPRMC,,V,,,,,,,091001,19.9,E,N*00
// 68bytes
// $GPRMC,000000,A,4916.45,S,12311.12,E,000.0,000.0,191194,020.3,E,N*00
//
// $GPRMC HHMMSS,<valid>,<lat>,<N/S>,<lon>,<E/W><speed><course>,DDMMYY,<mag>,<E/W>*<CS>
//
//   HHMMSS   = time in hours, min, sec -- UTC
//   <valid>  = A if data is valid, V if it is not (yea I know -- go figure)
//   <lat>    = latitude, eg 4916.45 = 49 degrees 16.45 min
//   <N/S>    = N for northern lat, S for southern lat
//   <lon>    = longitude, eg 12311.12 = 123 degrees 11.12 minutes
//   <E/W>    = E for east lon, W for west lon
//   <speed>  = speed over ground in knots
//   <course> = Course made good, True
//   DDMMYY   = date of fix. 191194 = 19 November 1994
//   <mag>    = magnetic variation, eg 020.3 = 20.3 degrees
//   <E/W     = direction of magnetic correction (east or west)
//   *<CS>    = check sum
//
//

   //Get time from GPS
   readGPS();

   //First get the time
   curtime = (unsigned int)((gpsdata[4] & 0x0F)
	   + (unsigned int)((gpsdata[3] & 0x0F)*6)
	   + (unsigned int)((gpsdata[2] & 0x0F)*60)
	   + (unsigned int)((gpsdata[1] & 0x0F)*360));

   // Now see if data is valid
   readGPS();	// get the <A/V> character

   // V = not valid.  Gota love it.
   if (gpsdata[0] == 'V')
    {
     valid=0;	// Kill the valid light and try again
     validpos = 0;
     continue; //Loop from start
    }// if

   break; //Exit loop

  } //End data loop

 // OK, we have good data.  Be sure the valid light is on (and remember)
 valid=1;
 validpos = 1;

#ifdef MIC_E
 info[0] = 0x60; //True reading
#endif //MIC_E

// OK, now read the position info and format the APRS message

 // Read the Latitude
 readGPS();

#ifdef MIC_E
 //Compress for MIC-E format.
 dest[0] = 64 | gpsdata[0];	//insert the first message bit
 dest[1] = 64 | gpsdata[1];  	//and the second message bit
 dest[2] = gpsdata[2]; 		//and the third message bit
 dest[3] = gpsdata[3];		//insert rest of the Latitude
 dest[4] = gpsdata[5];
 dest[5] = gpsdata[6];
#else //MIC_E
 //Insert data for non MIC-E format.
 for (i=0;i<7;i++) lat[i] = gpsdata[i];
#endif //MIC_E

 // Read the N/S byte and pack
 readGPS();

#ifdef MIC_E
 //Compress for MIC-E format.
 if (gpsdata[0] == 'N')
  dest[3] |= 64;
#else //MIC_E
 //Insert data for non MIC-E format.
 lat[7] = gpsdata[0];
#endif //MIC_E

 // Read the Longitude
 readGPS();

#ifdef MIC_E
 //Compress for MIC-E format.
 if (gpsdata[0] == '1')
  dest[4] |= 64; //insert first digit of Longitude

 //Pack the remaining portion of the Longitude
 info[1] = ((gpsdata[1] & 15)*10) + (gpsdata[2] & 15);

 //adjust 'D' if last two digits are less than 10
 if (info[1] < 10)
  {

   if (gpsdata[0] == '1') (info[1] += 80);   //adjustment for D = 100-109
   if (gpsdata[0] == '0')
    {
     //adjustment for D = 0-9  add 190 by adding 90 here and setting 100's digit
     info[1] += 90;
     dest[4] |= 64;
    }
  }

 info[1] += 28;
 if (info[2] <10) (info[2] += 60);  // adjust 'M' if its less than 10
 info[2] = ((gpsdata[3] & 15)*10) + (gpsdata[4] & 15) + 28;
 info[3] = ((gpsdata[6] & 15)*10) + (gpsdata[7] & 15) + 28;
#else //MIC_E
 //Insert data for non MIC-E format.
 for (i=0;i<8;i++) lon[i] = gpsdata[i];
#endif //MIC_E

 // read the E/W byte
 readGPS();

#ifdef MIC_E
 //Compress for MIC-E format.
 if (gpsdata[0] == 'W')
  dest[5] |= 64;

 //Set SSID to -0
 dest[6] = 0x60;
#else //MIC_E
 //Insert data for non MIC-E format.
 lon[8] = gpsdata[0];
#endif //MIC_E

// Now pack the course/speed portion of the message.

 // First the speed data
 getGPSfloat();

#ifdef MIC_E
 //Compress for MIC-E format.
 info[4] = (cs[0]*10) + cs[1];
 info[5] = (cs[2]*10);
#else //MIC_E
 //Insert data for non MIC-E format.
 cs[4] = cs[0];  // into cs[4] - cs[6]
 cs[5] = cs[1];
 cs[6] = cs[2];

 cs[3] = '/'; // part of the APRS spec.
#endif //MIC_E

 // Read the course info and pack
 // For non MIC-E format this is all needed
 getGPSfloat(); // into cs[0] - cs[2]

#ifdef MIC_E
 //Compress for MIC-E format.
 info[5] += cs[0];
 info[6] = (cs[1]*10) + cs[2];

 info[4] += 108;
 info[5] += 32;
 info[6] += 28;

 //adjustment as per mic-e docs.  Ensures all caps in the dest.
 for (i= 0;i<6;i++) dest[i] = toupper(dest[i]);

 //Callsign part of dest must be shifted left by 1
 //Doing it this way is more code efficient then the method it was done by in
 //the original code.
 //This way the LSB is always 0!
 dest[0] *= 2;
 dest[1] *= 2;
 dest[2] *= 2;
 dest[3] *= 2;
 dest[4] *= 2;
 dest[5] *= 2;
#endif //MIC_E

}//getdata


///////////////////////////////////////////////
//                                           //
//          crcbit(byte tbyte)()             //
//                                           //
///////////////////////////////////////////////
// Calc the packet CRC (check sum)
// See  "http://www.dot.ca.gov/hq/
// traffops/elecsys/ab3418/fcs-calc.html"
// for a good discussion on this
//
// This routine shamelessly lifted from
// W2FS' code
//
void crcbit(unsigned char tbyte)
{

//rotates the entire 16 bits to the right
#asm
 BCF    03,0
 RRF    _gpsdata+7,F
 RRF    _gpsdata+6,F
#endasm
 if (((STATUS & 0x01)^(tbyte)) ==0x01)
  {
   crchi = crchi^0x84;
   crclo = crclo^0x08;
  }// if
}// crcbit

///////////////////////////////////////////////
//                                           //
//          sendbyte(byte inbyte)()          //
//                                           //
///////////////////////////////////////////////
//
// Transmit the next byte in the packet
//
void sendbyte (unsigned char inbyte)
{

// int k, bt;
 unsigned char k;
 static bit bt;

 // Send 8 bits
 for (k=0;k<8;k++)
  {

   // overlap bit extraction and CRC calc of this bit with generation of
   // tone for last bit
   bt = inbyte & 0x01; //strip off the rightmost bit

   //do crc, but only if this is not a flag or crc byte
   if ( !crcflag && !flag) crcbit((unsigned char)bt);

//
// need to stop and sync up now.  Wait for last bit to finish
//
   if(doing1200)
    while( loop_ctr < loop12 ); //1.2KHz length
   else
    while( loop_ctr < loop22 ); //2.2KHz length

   loop_ctr = 0;

   if (!bt)			// if the next data bit to send is a zero
    flipout();			// then flip the output state
   else
    {
     stuff++;			// increment 1's count
     if (!flag && (stuff == 5))	// if 5 1s in a row (and not a CRC or flag)
      {
       if(doing1200)
	while( loop_ctr < loop12 );
       else
	while( loop_ctr < loop22 );
       loop_ctr = 0;
       flipout();		// then slip in a zero (X25 spec)
      }
    }// else
   // Rotate byte
#asm
   rrf   ?a_sendbyte,F
#endasm

  }// for
} // sendbyte


///////////////////////////////////////////////
//                                           //
//           sendpacket(void)()              //
//                                           //
///////////////////////////////////////////////
//
// All the stuff to send out a formatted APRS packet
//
//
// Packet format is:
//
// <flags> <to> <from> 03 F0 ! <lat> <table> <lon> <sym> <C/S> <text> <CRC> <flags>
//
void sendpacket(void)
{

 unsigned char i,j, beaconRate;

 // Update our configuration per switches
 if (!sw1)
  {
   beaconRate = eeprom_read(EEbrate1);            // config 1 beacon rate
#ifdef MIC_E
   info[7] = eeprom_read(EEsym1);  //set the Icon Symbol
   info[8] = eeprom_read(EEsymbtab1);  //set the Icon Table
#else //MIC_E
   sym = eeprom_read(EEsym1);       // config 1 symbol ID
   tab = eeprom_read(EEsymbtab1);   // config 1 symbol table ID
#endif //MIC_E
  }
 else
  {
   beaconRate = eeprom_read(EEbrate2);            // config 2 beacon rate
#ifdef MIC_E
   info[7] = eeprom_read(EEsym2);  //set the Icon Symbol
   info[8] = eeprom_read(EEsymbtab2);  //set the Icon Table
#else //MIC_E
   sym = eeprom_read(EEsym2);       // config 2 symbol ID
   tab = eeprom_read(EEsymbtab2);   // config 2 symbol table ID
#endif //MIC_E
  }

 //Configure tones
 initTone();

 //Configure CRC
 crclo = crchi = 0xFF;
 //Reset bit stuffing flag
 stuff = 0;
 //Reset CRC flag
 crcflag = 0;

 //If in beacon mode load the TxDelay.
 //If in MIC-E mode do a low fixed value.
 if ( sw2 )
  i = eeprom_read(EExdelay); // key-up delay
 else
  i = 16; //Should be enough for most TNCs

 // Keyup
 ptt_out=1;

 // Load timer for first trigger
 TMR0 = 0xff;
 //enable timer interrupt
 T0IE = 1;

 // Signal sending flags for CRC
 flag = 1;

 // Send flags during the key-up delay
 do { sendbyte(0x7e); } while (--i);

 // Done sending flags
 flag = 0;

#ifdef MIC_E
 for(i=0;i<7;i++) sendbyte(dest[i]);        //send destination bytes
#else //MIC_E
 // now send too address (which is really a software ID)
 for (i=0; i<7; i++) sendbyte(lookup[i]); //56
#endif //MIC_E

 // now send from and path info. Up to 3 paths are allowed
 i=EEfrom;

 do // + (3*7)
  {
   T0IE = 0;			// Timer interrupt off or the interrupt routines gets real confused!
   j = eeprom_read(i);		// Get byte from EEPROM
   T0IE = 1;			// Timer interrupt on
   sendbyte(j);			// Send it
   i++;				// Next byte in EEPROM
  } while (!(j & 0x01));	// loop until we find a byte with LSB set!

 // required protocol bytes
 sendbyte(0x03); //8
 sendbyte(0xf0); //8

#ifdef MIC_E
 for(i=0; i<9; i++) sendbyte(info[i]);    //send information part of packet
#else //MIC_E
 // No timestamp
 sendbyte('!'); //8

 // latitude
 for (i=0; i<8; i++) sendbyte(lat[i]); //56

 // Symbol table ID (/ or \)
 sendbyte(tab); //8

 // Longitude
 for (i=0; i<9; i++) sendbyte(lon[i]); //64

 // symbol
 sendbyte(sym); //8

 // Course and Speed
 for (i=0; i<7; i++) sendbyte(cs[i]); //48

#endif //MIC_E

 // then beacon if time
 // Only beacon if not MIC-E
 if ( sw2 )
  {

   if (--beacon == 0)
    {

     sendbyte(' ');			// First byte does not show up on D7a (this normal?)
     i=EEcomment;			// Start of the comment in EEPROM
     do
      {
       T0IE = 0;			// Timer off
       j = eeprom_read(i);		// Get byte
       T0IE = 1;			// Timer On
       if(j)sendbyte(j);		// Send it
       i++;				// Next byte
      } while (j != 0);			// Get bytes till 0 found
     beacon = beaconRate;		// Reset the beacon counter back to its base value
    } // if
  }

// finish with the CRC
 crcflag = 1;		// Signal sending CRC
 crclo ^= 0xff;		// Invert CRC
 crchi ^= 0xff;
 sendbyte(crclo);	//send the low byte of crc 8
 sendbyte(crchi);	//send the high byte of crc 8

// and two flags
 flag = 1;			// Signal sending flags
 sendbyte(0x7e);sendbyte(0x7e);	// Send flags to end packet 16
 flag = 0;			// Done flags

 T0IE = 0;			// Timer Off

 ptt_out=0;			// unkey PTT

} // sendpacket

#ifdef MORSE
//Send morse ID every 15mins.
//Will only be active in in beacon mode.
//Dot clock is 50mS
void send_morse_id(void)
{

 unsigned char i,j,k,l,m;

 //How long since last?
 morse_t = curtime - lastmorse;

 //Has 15min elasped?
 if ( morse_t < 90 ) return;

 //Send it
 ptt_out = 1; //Key up

 i = EEfrom; //Start of the callsign in EEPROM

 m = 0; //Count of chars sent

 for(;;)
  {

   j = (eeprom_read(i)/2); // Get byte from EEPROM & shit right by one bit

   //If found a space or done 6 chars then quit
   if ( ( j == ' ' ) || ( m == 6 ) ) break;

   /* Check for a number */
   k = j - '0';

   /* Is it a number? */
   if ( k < 10 )
    j = n_tabel[k]; //Is a number
   else
    {

     /* Action a char */
     k = j - 'A';

     /* It must be a letter? */
     j = a_tabel[k];

    }

   //Get bit count
   k = j & 7;

   //Do it..
   do
    {

     l = 0; //Used for tone length

     //Configure tones, using the 1.2KHz tone
     initTone();

     // Load timer for first trigger
     TMR0 = 0xff;
     //enable timer interrupt
     T0IE = 1;

     //Check for a . or -
     if ( j & 0x80 )
      {

       //Do a dash
       for(;;)
	{

	 while(loop_ctr < 39 ); //~1ms of tone
	 loop_ctr = 0; //Reset loop counter
	 l++; //Done 1ms of tone
	 //Done 150ms of tone?
	 if ( l == 150 )
	  {

	   T0IE = 0; //Done 150ms

	   break;

	  }
	}
      }
     else
      {

       //Do a dot
       for(;;)
	{

	 while(loop_ctr < 39 ); //~1mS of tone
	 loop_ctr = 0; //reset loop counter
	 l++; //Done 1ms of tone
	 //Done 50ms of tone?
	 if ( l == 50 )
	  {

	   T0IE = 0; //Done 50ms

	   break;

	  }
	}
      }

     //Pause 50mS between elements
     //ONLY if not last element!
     if(k != 1)delay_ms(50);
     j <<= 1; //Next element

    } while(--k); //End while

   delay_ms(150); //Pause 150mS between chars

   i++; // Next byte in EEPROM
   m++; //Done a char

  } //End for

// ptt_out = 0; //De-key
 lastmorse = curtime; //Load time for next beacon

}
#endif //MORSE

///////////////////////////////////////////////
//                                           //
//                    MAIN                   //
//                                           //
///////////////////////////////////////////////
void main(void)
{

//
//
//                       RA4    RA3    RA2    RA1    RA0
//                       In     Out    Out    Out    Out
//                       Sin    So     PTT    Val    CD
//   0      0      0      1      0      0      0      0  = 0x10
//
//  RB7    RB6    RB5    RB4    RB3    RB2    RB1    RB0
//  In     In     In     In     Out    Out    Out    Out
//  SW2    SW1    PTT    Ain    R1K    R2K    R4K    R8K
//   1      1      1      1      0      0      0      0  = 0xF0
 TRISA = 0x10;
 TRISB = 0xf0;

 //Enable PORTB pull ups and set prescale /1
 OPTION = 8;

 //Zero volts at the resistor network and No PTT
 PORTB = 0;

 //CD LED on, VALID & PTT LEDs off and serial line low
#ifndef DEBUG
 PORTA = 1;
#else
 PORTA = 0;
#endif

 //Tell the world we are OK

#ifndef DEBUG
 //Pause for 0.25sec
 delay_ms(250);
#endif
 //LED off
 cd=0;
#ifndef DEBUG
 //Pause for 0.25sec
 delay_ms(250);
#endif
 //LED on
 cd=1;
#ifndef DEBUG
 //Pause for 0.25sec
 delay_ms(250);
#endif
 //LED off
 cd=0;

 //Make sure if no GPS is connected the MIC-E produces valid X25 data
#ifdef MIC_E
 dest[0] = lookup[0];
 dest[1] = lookup[1];
 dest[2] = lookup[2];
 dest[3] = lookup[3];
 dest[4] = lookup[4];
 dest[5] = lookup[5];
 dest[6] = lookup[6];
 info[0] = '>';
 info[1] = 'N';
 info[2] = 'o';
 info[3] = ' ';
 info[4] = 'G';
 info[5] = 'P';
 info[6] = 'S';
#else //MIC_E
 lat[0] = '>';
 lat[1] = 'N';
 lat[2] = 'o';
 lat[3] = ' ';
 lat[4] = 'G';
 lat[5] = 'P';
 lat[6] = 'S';
#endif //MIC_E

 //Configure flags.
 init();

#ifdef TIMED_MIC_E
 //Signal OK to send. Used for the No GPS message.
 beacon_ok = 1;
#endif

 //Start interrupts
 ei();

 // Main loop ... do this until power is off or reset
 for(;;)
  {

   // get position and update curtime (# of 10 second chunks that have gone by)
   // The only way to return is with valid position data
   // or if send_mice is set.
   getdata();

   // need to do this here. getdata will bonk xmitrate
   if (!sw1)
     xmitrate = eeprom_read(EEtrate1);
   else
     xmitrate = eeprom_read(EEtrate2);

#ifndef TIMED_MIC_E
   //Check for MIC-E mode
   if ( !sw2 )
    {

     //Need to action mic-e?
     if ( send_mice )
      {

       send_mice = 0; //Clear the flag

#ifndef DEBUG
       //Do a debounce
       delay_ms(100);
#endif

       //Still De-keyed?
       if (ptt_in)
	{

	 //Send data
	 //Will be either the No GPS message or last valid GPS data
	 sendpacket();

	 ptt_active = 0; //Signal ptt not active

	} //End ptt_in
      } //End send_mice

     continue; //Ignore all else

    } //End sw2
#endif //TIMED_MIC_E

   // First time in the loop?
   // If so skip the time check
   if (!firstTime)
    {

     //How long since last beacon?
     delta = curtime - lasttime;

     //Need to do a packet beacon?
     if (delta < xmitrate)
#ifndef TIMED_MIC_E
      continue;
#else //TIMED_MIC_E
      beacon_ok = 0;
     else
      beacon_ok = 1;
#endif //TIMED_MIC_E
    }// if

#ifdef TIMED_MIC_E
   //Check for MIC-E mode
   if ( !sw2 )
    {

     //Need to action mic-e?
     if ( send_mice )
      {

       send_mice = 0; //Clear the flag

#ifndef DEBUG
       //Do a debounce
       delay_ms(100);
#endif

       //Still De-keyed?
       if (ptt_in)
	{

	 //Has valid data been seen from the GPS?
	 if (validpos)
	  {
	   //GPS has produced valid data, so check if OK too send a beacon
	   if (beacon_ok)
	    {
	     //Need too update the timer for next MIC-E beacon
	     firstTime = 0; //Not first time in the loop
	     lasttime = curtime; //Reload for next period
	     //Will ALWAYS be the current GPS position
	     sendpacket();
	     beacon_ok = 0; //reset beacon flag
	    }
	  }
	 else
	  //GPS has not produced valid data so just beacon
	  //Will be either the No GPS message or the LAST valid GPS position
	  sendpacket();

	 //ALWAYS need to De-Key due too the valid data check!
	 ptt_out = 0;
	 ptt_active = 0; //Signal ptt not active

	} //End ptt_in
      } //End send_mice

     continue; //Ignore all else

    } //End sw2

   //If no need to send a timed beacon then no further action
   if(!beacon_ok) continue;
#endif //TIMED_MIC_E

   // We fall through when it is time to send a packet in beacon mode
   if ( validpos )
    {
     firstTime = 0; //Not first time in the loop
     lasttime = curtime; //Reload for next period
    }

   // First wait until we see no traffic for qtime
   cqtime = qtime = eeprom_read(EEqtime);
   do
    {
#ifndef DEBUG
     delay_ms(100);  // wait 1/10 second
#endif
     flashcd();      // update the CD lamp
     if (cdflag) qtime = cqtime;
    } while (--qtime);

#ifdef MORSE
   //Check for and action the Morse ID.
   send_morse_id();
#endif

   // Then send the packet
   sendpacket();

// Byon communicated an important point here.  If the APRS channel is
// busy we could stay in this loop for a long time and the APRS data
// could get quite old.  One could loop here for some maximum time, say
// 20 seconds, then abort the transmission and go back for fresh GPS data.
//
// For example, we could define a variable loopCtr then say
//
//    cqtime = qtime = read_eeprom(EEqtime);
//*   loopCtr = 0;
//
//    do{
//       qtime--;
//       delay_ms(100);
//*      loopCtr++;
//       flashcd();
//      if (cdflag) (qtime = cqtime);
//       }
//*   while ((qtime != 0) && (loopCtr < 200)); // loop for a max of 20 seconds
//
//*   if (qtime==0) (sendpacket();  // send data if channel quite for qtime
//
//  This would wait for 20 seconds before giving up, trashing the
//  gps data, and trying again.
//
//  I have not yet tested this code so I leave it as a suggestion.
//

  }//end of while true
}//end of main
