#include <pic.h> //Needed
#include "sottpar.h" //EEPROM locations

//Only use for code debuging!
//#define DEBUG

/****************************************************************************

 If MIC_E is defined then packet data will be sent in MIC-E format.

 If MIC_E is not defined the packet data will be sent in the default SOTT
 format.

 This does not effect the ptt action!
*/
#define MIC_E

/****************************************************************************

 If TIMED_MIC_E is defined the the mic-e packets will only be send each
 intervile currently selected with sw1.

 Will stop a packet being sent on each over.
*/
//#define TIMED_MIC_E

/****************************************************************************

 If morse is defined then the units callsign will be sent at the end of each
 packet transmission.

 The morse WILL not be sent if the transmission was the result of the MIC-E
 ptt!.

 Cannot be used with the MIC_E define due to code space.

 HAS NOT BEEN TESTED!!
*/
//#define MORSE

#if defined(MORSE) && defined(MIC_E)
#undef MIC_E
#endif

#if defined(TIMED_MIC_E) && !defined(MIC_E)
#undef TIMED_MIC_E
#endif

// Define global constants
#define GPS_buffer_size 9
#define w_tmp     gpsdata[0]   // Save W register in interrupt loop
#define stat_tmp  gpsdata[1]   // Save status register in interrupt loop
#define r_val     gpsdata[2]   // Current value at resistor network
#define loop_ctr  gpsdata[3]   // Used by sendbyte to time bit length

#define xmitrate  gpsdata[4]   // 10s of seconds between pack transmissions

#define base_dink gpsdata[5]   // The value used to reinit dink
#define crclo     gpsdata[6]   // the CRC aka FCS or check bytes
#define crchi     gpsdata[7]   // the CRC aka FCS or check bytes
#define stuff     gpsdata[8]   // bit stuff counter

// Main use of gpsdata
#define qtime     gpsdata[0]   // Used by transmit holdoff in main
#define cqtime    gpsdata[1]   // cache qtime to avoid eeprom read

// Config use of gpsdata
#define Ccmd      gpsdata[0]   // Be careful, inittone will change
#define Cstart    gpsdata[1]   // some of the gpsdata and Config
#define Cend      gpsdata[2]   // calls inittone
#define Cloop     gpsdata[3]
#define Cdata     gpsdata[4]

#ifdef MAIN
#define EXTERN
// The "To" address below is an APRS software ID.  It must start with AP
// I use APTT25-0 for TinyTrak 2.5
const unsigned char lookup[7] = {('A'*2), ('P'*2), ('T'*2), ('T'*2), ('2'*2), ('5'*2), ('0'*2)};
#else
#define EXTERN extern
extern const unsigned char lookup[];
#endif

// RAM space is very tight so we will use gpsdata[n] in four ways.
//
//  (1) While getting data, gpsdata will buffer pieces of the GPS string.
//  (2) When xmitting gpsdata will be used to hold interrupt routine
//      values and data used by sendpacket (crc info, etc)
//  (3) In main, gpsdata will hold misc values.
//  (4) In Config, gpsdata will hold start/stop addresses, temp data, etc.
EXTERN unsigned char gpsdata[GPS_buffer_size]; // buffer for GPS data


// Xmit data phase use (Interrupt routine and sendpacket)


// These values are used by the interrupt routine and could go in
// gpsdata if there was space
EXTERN unsigned char dink;		// holds the current dink value.
EXTERN unsigned char base;		// current timr0 init value (defines nominal frequency)
EXTERN unsigned char cbase1200;	// cached value of base1200 (avoids a read_eeprom)
EXTERN unsigned char cbase2200;	// cached value of base2200 (avoids a read_eeprom)
EXTERN unsigned char loop12, loop22;

// The following variables are filled during get gps data phase and used
// during transmit
#ifdef MIC_E
EXTERN unsigned char dest[7];
EXTERN unsigned char info[9];
#else
EXTERN unsigned char lat[8];	// latitude from GPS
EXTERN unsigned char lon[9];	// longitude from GPS
#endif
EXTERN unsigned char cs[7];	// course and speed from GPS

// The following variables need to persist during all phases of the program
EXTERN unsigned int lasttime;	// Remembers the time when the last packet was transmitted.
EXTERN unsigned int curtime;
#ifdef MORSE
EXTERN unsigned int lastmorse;
EXTERN unsigned char morse_t;
#endif
EXTERN unsigned char beacon;	// counts the # of xmits to know when to send comment
EXTERN unsigned char delta;

// Used for symbol and tabel
EXTERN unsigned char sym, tab;

// Misc flags
EXTERN bit flag, crcflag, cdflag, validpos, firstTime, doing1200;
EXTERN bit send_mice, ptt_active, beacon_ok, d_string;
// flag       => Are we sending flags on RF?    1 = yes
// crcflag    => Are we sending CRC data on RF? 1 = yes
// cdflag     => Carrier Detect (is RF busy)?   1 = yes
// valispos   => Is the GPS data valid?         1 = yes
// firstTime  => Is this our first packet?      1 = yes
// doing1200  => Is the current tone 1200Hz?    1 = yes
// ptt_active => Used for MIC-E.
// send_mice  => Flag to action MIC-E PTT control.
// beacon_ok  => Set when beacon time has elapsed.
// d_string   => Signals reading RMC or GGA string from GPS

#ifdef DEBUG
EXTERN unsigned char debug_c;
#endif

//Protos
void delay_ms(unsigned char cnt);
unsigned char getch(void);
void putch(unsigned char);
